# coding:UTF-8
'''
对原始的 eeg 信号，使用 CNN 进行情感分类。
网络结构为：3层CNN+2层FC
'''

import torch
import torch.nn as nn
import torch.nn.functional as F


# 定义卷积网络结构
class ConvNet(nn.Module):
    def __init__(self, num_classes):
        super(ConvNet, self).__init__()
        self.layer1 = nn.Sequential(
            nn.Conv2d(1, 32, kernel_size=5, stride=(1, 1), padding=(2, 2), bias=True),
            nn.BatchNorm2d(32),
            nn.LeakyReLU(),
            nn.AvgPool2d(kernel_size=2, stride=2)
        )
        self.layer2 = nn.Sequential(
            nn.Conv2d(32, 64, kernel_size=5, stride=(1, 1), padding=(2, 2), bias=True),
            nn.BatchNorm2d(64),
            nn.LeakyReLU(),
            nn.AvgPool2d(kernel_size=2, stride=2)
        )
        self.layer3 = nn.Sequential(
            nn.Conv2d(64, 128, kernel_size=3, stride=(1, 1), padding=(1, 1), bias=True),
            nn.BatchNorm2d(128),
            nn.LeakyReLU(),
            nn.AvgPool2d(kernel_size=2, stride=2)
        )
        self.fc1 = nn.Linear(128 * 7 * 25, 256, bias=True)
        self.fc2 = nn.Linear(256, num_classes, bias=True)

    def forward(self, x):
        out = self.layer1(x)
        out = self.layer2(out)
        out = self.layer3(out)
        out = out.reshape(out.size(0), -1)
        out = self.fc1(out)
        out = self.fc2(out)
        return out

#Example
'''
model = ConvNet(num_classes).to(device)

parameter:
num_classes:想要分成的情绪的种类


# Loss and optimizer
criterion = nn.CrossEntropyLoss()
optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate, weight_decay=0.001)
scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=3, gamma=0.95)

'''


class ScaledDotProductAttention(nn.Module):
    ''' Scaled Dot-Product Attention '''

    def __init__(self, temperature, attn_dropout=0.1):
        super().__init__()
        self.temperature = temperature
        self.dropout = nn.Dropout(attn_dropout)

    def forward(self, q, k, v, mask=None):
        attn = torch.matmul(q / self.temperature, k.transpose(2, 3))

        if mask is not None:
            attn = attn.masked_fill(mask == 0, -1e9)

        attn = self.dropout(F.softmax(attn, dim=-1))
        output = torch.matmul(attn, v)

        return output, attn


class MultiHeadAttention(nn.Module):
    ''' Multi-Head Attention module '''

    def __init__(self, n_head, d_model, d_k, d_v, dropout=0.1):
        super().__init__()

        self.n_head = n_head
        self.d_k = d_k
        self.d_v = d_v

        self.w_qs = nn.Linear(d_model, n_head * d_k, bias=False)
        self.w_ks = nn.Linear(d_model, n_head * d_k, bias=False)
        self.w_vs = nn.Linear(d_model, n_head * d_v, bias=False)
        self.fc = nn.Linear(n_head * d_v, d_model, bias=False)

        self.attention = ScaledDotProductAttention(temperature=d_k ** 0.5)

        self.dropout = nn.Dropout(dropout)
        self.layer_norm = nn.LayerNorm(d_model, eps=1e-6)

    def forward(self, q, k, v, mask=None):
        d_k, d_v, n_head = self.d_k, self.d_v, self.n_head
        sz_b, len_q, len_k, len_v = q.size(0), q.size(1), k.size(1), v.size(1)

        residual = q

        # Pass through the pre-attention projection: b x lq x (n*dv)
        # Separate different heads: b x lq x n x dv
        q = self.w_qs(q).view(sz_b, len_q, n_head, d_k)
        k = self.w_ks(k).view(sz_b, len_k, n_head, d_k)
        v = self.w_vs(v).view(sz_b, len_v, n_head, d_v)

        # Transpose for attention dot product: b x n x lq x dv
        q, k, v = q.transpose(1, 2), k.transpose(1, 2), v.transpose(1, 2)

        if mask is not None:
            mask = mask.unsqueeze(1)  # For head axis broadcasting.

        q, attn = self.attention(q, k, v, mask=mask)

        # Transpose to move the head dimension back: b x lq x n x dv
        # Combine the last two dimensions to concatenate all the heads together: b x lq x (n*dv)
        q = q.transpose(1, 2).contiguous().view(sz_b, len_q, -1)
        q = self.dropout(self.fc(q))
        q += residual

        q = self.layer_norm(q)

        return q, attn


class PositionwiseFeedForward(nn.Module):
    ''' A two-feed-forward-layer module '''

    def __init__(self, d_in, d_hid, dropout=0.1):
        super().__init__()
        self.w_1 = nn.Linear(d_in, d_hid)  # position-wise
        self.w_2 = nn.Linear(d_hid, d_in)  # position-wise
        self.layer_norm = nn.LayerNorm(d_in, eps=1e-6)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        residual = x

        x = self.w_2(F.relu(self.w_1(x)))
        x = self.dropout(x)
        x += residual

        x = self.layer_norm(x)

        return x


def get_pad_mask(seq, pad_idx):
    return (seq != pad_idx).unsqueeze(-2)


def get_subsequent_mask(seq):
    ''' For masking out the subsequent info. '''
    sz_b, len_s = seq.size()
    subsequent_mask = (1 - torch.triu(
        torch.ones((1, len_s, len_s), device=seq.device), diagonal=1)).bool()
    return subsequent_mask


class PositionalEncoding(nn.Module):

    def __init__(self, d_hid, n_position=200):
        super(PositionalEncoding, self).__init__()

        # Not a parameter
        self.register_buffer('pos_table', self._get_sinusoid_encoding_table(n_position, d_hid))

    def _get_sinusoid_encoding_table(self, n_position, d_hid):
        ''' Sinusoid position encoding table '''

        # TODO: make it with torch instead of numpy

        def get_position_angle_vec(position):
            return [position / np.power(10000, 2 * (hid_j // 2) / d_hid) for hid_j in range(d_hid)]

        sinusoid_table = np.array([get_position_angle_vec(pos_i) for pos_i in range(n_position)])
        sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2])  # dim 2i
        sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])  # dim 2i+1

        return torch.FloatTensor(sinusoid_table).unsqueeze(0)

    def forward(self, x):
        return x + self.pos_table[:, :x.size(1)].clone().detach()


class EncoderLayer(nn.Module):
    ''' Compose with two layers '''

    def __init__(self, d_model, d_inner, n_head, d_k, d_v, dropout=0.1):
        super(EncoderLayer, self).__init__()
        self.slf_attn = MultiHeadAttention(n_head, d_model, d_k, d_v, dropout=dropout)
        self.pos_ffn = PositionwiseFeedForward(d_model, d_inner, dropout=dropout)

    def forward(self, enc_input, slf_attn_mask=None):
        enc_output, enc_slf_attn = self.slf_attn(
            enc_input, enc_input, enc_input, mask=slf_attn_mask)
        enc_output = self.pos_ffn(enc_output)
        return enc_output, enc_slf_attn


class Encoder(nn.Module):
    ''' A encoder model with self attention mechanism. '''

    def __init__(
            self, d_word_vec, n_layers, n_head, d_k, d_v,
            d_model, d_inner, dropout=0.1, n_position=200):

        super().__init__()

        # self.src_word_emb = nn.Embedding(n_src_vocab, d_word_vec, padding_idx=pad_idx)
        self.position_enc = PositionalEncoding(d_word_vec, n_position=n_position)
        self.dropout = nn.Dropout(p=dropout)
        self.layer_stack = nn.ModuleList([
            EncoderLayer(d_model, d_inner, n_head, d_k, d_v, dropout=dropout)
            for _ in range(n_layers)])
        self.layer_norm = nn.LayerNorm(d_model, eps=1e-6)

    def forward(self, src_seq, src_mask, return_attns=False):
        enc_slf_attn_list = []

        # -- Forward

        enc_output = self.dropout(self.position_enc(src_seq))
        enc_output = self.layer_norm(enc_output)

        for enc_layer in self.layer_stack:
            enc_output, enc_slf_attn = enc_layer(enc_output, slf_attn_mask=src_mask)
            enc_slf_attn_list += [enc_slf_attn] if return_attns else []

        if return_attns:
            return enc_output, enc_slf_attn_list
        return enc_output,


class CorruptionLayer(nn.Module):
    def __init__(self, device, corrupt_probability=0.1):
        super(CorruptionLayer, self).__init__()
        self.corrupt_p = corrupt_probability
        self.device = device

    def forward(self, feature):
        bitmask = torch.cuda.FloatTensor(feature.shape).uniform_() > self.corrupt_p
        return torch.mul(feature, bitmask)


class TransformerEncoder(nn.Module):
    """TransformerEncoder is a stack of N encoder layers.

    Args:
        sentence_len(int)   : number of samples for each input
        seq_len    (int)    : length of each sequence
        n_layers   (int)    : number of sub-encoder-layers in the encoder
        n_heads    (int)    : number of heads in the multiheadattention models
        p_drop     (float)  : dropout value
        d_ff       (int)    : dimension of the feedforward network model

    """

    def __init__(self, sentence_len, d_feature, n_layers=6, n_heads=8, p_drop=0.1, d_ff=2048):
        super(TransformerEncoder, self).__init__()
        d_k = d_v = d_feature // n_heads
        self.encoder = Encoder(n_position=sentence_len,
                               d_word_vec=d_feature, d_model=d_feature, d_inner=d_ff,
                               n_layers=n_layers, n_head=n_heads, d_k=d_k, d_v=d_v,
                               dropout=p_drop)
        self.linear = nn.Linear(d_feature, 3)
        self.softmax = nn.Softmax(dim=-1)

    def forward(self, src_seq):
        # src_mask = get_pad_mask(src_seq, self.src_pad_idx)
        src_mask = None
        outputs_feature, *_ = self.encoder(src_seq, src_mask)
        outputs, _ = torch.max(outputs_feature, dim=1)
        # |outputs| : (batch_size, d_model)
        outputs_classification = self.softmax(self.linear(outputs))
        return outputs_feature, outputs_classification




import scipy.io as sio
import numpy as np
import matplotlib.pyplot as plt
# from process import denoise
# from deeg.check import check_nan
# from deeg.band import band
# from deeg.sampling import sampling
# from deeg.segment import segment
from deeg.features import cal_eeg_features

data_dir = "D:/EEG/data_preprocessed_matlab/"

s01 = sio.loadmat(data_dir+"s01.mat")
s01_data, s01_label = s01["data"], s01["labels"]

single_s01_data = s01_data[0,0,:]
print(s01_data.shape, single_s01_data.shape)
print(len(s01_data.shape), len(single_s01_data.shape))

# ----- Extract all features -----
eeg_features, feature_list = cal_eeg_features(s01_data, 1000, 2000)
print(eeg_features.shape)
# from deeg.features import power_spectral_density
# a = power_spectral_density(single_s01_data, 1000, 2000)
# print(a)


# ----- Process: segment -----
# segmented_array = segment(single_s01_data, length=10, overlap=3)
#
# print(segmented_array[0].shape, segmented_array[1].shape)
# print(single_s01_data[:20])
# print(segmented_array.shape)




# ----- Process: sampling -----
# sampled_array = sampling(s01_data, 10, 1000)
# print(sampled_array.shape)

# ----- Process: band -----
# ts = band(single_s01_data)
# print(ts)
# print(ts["alpha"].shape)

# ----- Process: check -----
# print(check_nan(s01_data))

# ----- Features: DE -----
# def cal_de(s):
#     return (1 / 2) * np.log(2 * np.pi * np.e * np.std(s) ** 2)
#
# print(cal_de(single_s01_data))
#
#
# b = np.ones((2,3,2))
# c = np.apply_along_axis(cal_de, axis=-1, arr=s01_data)
# print(c.shape)
# print(c)
# print(cal_de(s01_data[39,39]))
# -----




import numpy as np
import re
from scipy import io
from scipy.io import loadmat
from numpy import genfromtxt


# ======Utils for dataset======
"""
Utils for loading DEAP Dataset
"""
def data_partition(data):
    X = data['data']
    y = data['labels']
    return X,y

def data_integration(data_list):
    data, labels = data_partition(io.loadmat(data_list[0]))
    for i in range(1, len(data_list)):
        X,y = data_partition(io.loadmat(data_list[i]))
        data = np.concatenate((data, X), axis = 0)
        labels = np.concatenate((labels, y), axis = 0)
    print("Shape of data: {}".format(data.shape))
    print("Shape of labels: {}".format(labels.shape))
    return data, labels

"""
Utils for loading NeuroMarketing Dataset
"""
def tryint(s):
    try:
        return int(s)
    except ValueError:
        return s

def str2int(v_str):
    return [tryint(sub_str) for sub_str in re.split('([0-9]+)', v_str)]

def sort_NeuroMarketing(v_list):
    return sorted(v_list, key=str2int)

"""
Utils for loading SEED Dataset
"""
def get_frequency_band_idx(frequency_band):
    lookup = {'delta': 0,
              'theta': 1,
              'alpha': 2,
              'beta': 3,
              'gamma': 4}
    return lookup[frequency_band]

def get_labels(label_path):
    return loadmat(label_path, verify_compressed_data_integrity=False)['label'][0]

"""
Utils for generic loading
"""
def read_csv(fname):
    return genfromtxt(fname, delimiter=",")




import numpy as np
from scipy import signal


def data_quality_check(array):
    """
    Check whether the input signal contains NaN, print relevant information
    input:
        array: numpy array. The temporal signal with 1d or with multiple dimensions
    return:
        None or dict. Dict stores indicator of Nan/Abnormal values.
            If there is no NaN in signal, return None, else return the index where NaN occurs
    """
    print("*** Data Quality Check ***")

    def check_nan(array):
        return np.isnan(array).any()

    def check_abnormal(array):
        """
        Todo: Define what kinds of EEG data are abnormal in some kinds?
        """
        return None

    if not check_nan(array) and check_abnormal(array):
        print("*** Data Quality Check Passed ***")
        return None
    else:
        # return format: tuple(array([a,b,c,...]), array([j,k,l,...]), ...)
        # Each element in tuple indicates the index list
        dict = {}
        if check_nan(array):
            print("Nan Values appear in data")
            print("The percentage of Nan Values is {val}%").format(val=sum(np.isnan(array)) / array.shape[0])
            print("The location is at {tuple}").format(tuple=np.where(np.isnan(array)))
            dict["nan"] = np.where(np.isnan(array))
        if check_abnormal(array):
            #Todo: add abnormal values' locations to dict.
            pass
        return dict


def savgol(array, window_size=5, polyorder=2):
    """
    Apply the Savitzky-Golay fiklter to smooth signals.
    We support both 1d array input and DEEG standard input (p,m,e).
    For the later format, we calculate DE on dim "e" and return a (p,m) numpy array.

    Input
    -------
    array: numpy array or list
    window_size: int
        the length of the filter window (i.e., the number of coefficients).
        window_length` must be a positive odd integer.
        default is 5.
    polyorder : int
        the order of the polynomial used to fit the samples.
        polyorder must be less than window_length.
        default is 2.

    Return
    -------
    y :ndarray
        same shape as x.
        The filtered data.
    """
    def create_x(size, rank):
        # creat weighting coefficients
        x = []
        for i in range(2 * size + 1):
            m = i - size
            row = [m ** j for j in range(rank)]
            x.append(row)
        x = np.mat(x)
        return x

    def SG(array, window_size, polyorder):
        if window_size % 2 != 1:
            raise ValueError("window_size must be odd")
        if polyorder >= window_size:
            raise ValueError("polyorder must be less than window_lengthd")

        m = int((window_size - 1) / 2)
        odata = np.array(array).tolist()
        for i in range(m):
            odata.insert(0, odata[0])
            odata.insert(len(odata), odata[len(odata) - 1])

        # creat X matrix
        x = create_x(m, polyorder)
        # calculate the weighting coefficients
        b = (x * (x.T * x).I) * x.T
        a0 = b[m]
        a0 = a0.T
        # calculate the filtered signal
        filtered_signal = []
        for i in range(len(array)):
            y = [odata[i + j] for j in range(window_size)]
            y1 = np.mat(y) * a0
            y1 = float(y1)
            filtered_signal.append(y1)
        return np.array(filtered_signal)
    if len(np.array(array).shape) == 1:
        return SG(array, window_size, polyorder)
    else:
        return np.apply_along_axis(SG, axis=-1, arr=array, window_size=window_size, polyorder=polyorder)


def filter(array, name="delta", order=5):
    """
    A band-pass filter to suppress signal out of frequency band
    input:
        array: numpy array. The input temporal signal with 1d
        name: str. Specify the filter type commonly used in EEG analysis. (delta, theta, alpha, beta, gamma)
        order: int. The order of Butterworth filter. Default is 5
    return:
        ts: numpy array. Filtered signal in temporal domain
    """
    if name == "delta":
        band = [1,3]
    elif name == "theta":
        band = [4,7]
    elif name == "alpha":
        band = [8,13]
    elif name == "beta":
        band = [14,30]
    elif name == "gamma":
        band = [31,50]
    else:
        raise(Exception("Invalid filter name!"))
    sos = signal.butter(order, band, 'bp', fs=1000, output='sos')
    ts = signal.sosfilt(sos, array)
    return ts


def band(array):
    """
    Filter temporal signal by all the 5 filters commonly used in EEG analysis
    input:
        signal: numpy array. The input temporal signal. 1d or with multiple dimensions
    return
        ts_dict: dictionary. Keys are filter name and values are filtered signal in temporal domain.
    """
    ts_dict = {}
    if len(array.shape) < 2:
        for name in ["delta","theta","alpha","beta","gamma"]:
            ts_dict[name] = filter(array)
    else:
        for name in ["delta","theta","alpha","beta","gamma"]:
            ts_dict[name] = np.apply_along_axis(filter, -1, array, name)
    return ts_dict


def segment(array, length=10, overlap=1, label=None):
    """
    Generate the signal segments to augment sample size.
    input:
        array: numpy array. The input temporal signal. 1d or with multiple dimensions
        length: int. The length of segments. Default is 10, which means each signal segment contains 10 data points.
        overlap: int. The overlap length between neighboring segments. Default is 1.
        label: numpy array. If label is not None, duplicate labels according to the data
    return:
        segmented_array: numpy array with shape (p,m,s,e) or (s,e), depending on the input. s is the number of segments.
        duplicated_labels: numpy array with shape (p,m,s,e) or (s,e). If input label is None do not return.
    """
    if len(array.shape) < 2:
        segmented_array = [array[i:i+length] for i in range(0, len(array), length-overlap)]
        if len(segmented_array[-1]) != length:
            segmented_array = segmented_array[:-1]
        if label is None:
            return np.array(segmented_array)
        else:
            return np.array(segmented_array), np.expand_dims(label, 0).repeat(len(segmented_array), axis=0)
    else:
        segmented_array = [array[:,:,i:i+length] for i in range(0, array.shape[-1], length-overlap)]
        if segmented_array[-1].shape[-1] != length:
            segmented_array = segmented_array[:-1]
        if label is None:
            return np.array(segmented_array).transpose(1,2,0,3)   # transpose numpy array to make (p,m,s,e) format
        else:
            return np.array(segmented_array).transpose(1,2,0,3), np.expand_dims(label, -2).repeat(segmented_array.shape[-2], axis=-2)


def sampling(array, interval=1, offset=0):
    """
    Down-sample the input signal with certain interval.
    input:
        array: numpy array. The input temporal signal. 1d or with multiple dimensions
        interval: int. The interval to sample EEG signal. Default is 1, which means NO down-sampling is applied
        offset: int. Sampling starts from "offset-th" data point
    return:
        sampled_array: numpy array. Down-sampled signal
    """
    if len(array.shape) < 2:
        return array[offset::interval]
    else:
        return array[:, :, offset::interval]


"""
Main package for loading common data file types:
    1. csv
    2. Common MNE-supported file types (txt, mat, etc.)
to numpy array with dimension (p, m, e).
"""
import os
import re
import numpy as np
from scipy.io import loadmat
from mne.io import read_raw
from pathlib import Path
from .utils import read_csv,data_integration, sort_NeuroMarketing, get_labels, get_frequency_band_idx

supported = {
    ".edf": read_raw,
    ".bdf": read_raw,
    ".gdf": read_raw,
    ".vhdr": read_raw,
    ".fif": read_raw,
    ".fif.gz": read_raw,
    ".set": read_raw,
    ".cnt": read_raw,
    ".mff": read_raw,
    ".nxe": read_raw,
    ".hdr": read_raw,
    ".mat": read_raw,
    ".bin": read_raw,
    ".data": read_raw,
    ".sqd": read_raw,
    ".con": read_raw,
    ".ds": read_raw,
    ".txt": read_raw,
    ".csv": read_csv
}

def load_data(fname, *, preload=False, verbose=None, **kwargs):
    """
    :param fname: Files you are gonna read
    :param mne: Indicate if the file is mne-supported
    :return: numpy array of eeg data
    """
    ext = "".join(Path(fname).suffixes)
    if ext in supported:
        if ext == ".csv":
            return supported[ext](fname)
        else:
            raw = supported[ext](fname, preload=preload, verbose=verbose, **kwargs)
            data, time = raw[:]
            return data

"""
Main package for loading existing public datasets:
    1. DEAP
    2. SEED
    3. DREAMER
    4. NeuroMarketing
to numpy array with dimension (p, m, e).
"""
def load_DEAP(path):
    """
    :param path: path to DEAP dataset
    DEAP dataset contains 32 participant files so the path points to the folder that stores those files.
    :return: numpy array of data and labels
    """
    if path[-1] != "/":
        path += "/"
    mat_list_1 = [path + 's0' + str(x) for x in range(1, 9)]
    mat_list_2 = [path + 's' + str(x) for x in range(10, 33)]
    data, labels = data_integration(mat_list_1 + mat_list_2)
    return data, labels

def load_NeuroMarketing(path):
    """
    :param path: path to NeuroMarketing dataset
    :return: numpy array of data and labels
    """
    original_name = []
    for file in os.listdir(path + '25-users/'):
        if re.search('\.txt$', file):
            original_name.append(file[:-4])
    name = sort_NeuroMarketing(original_name)
    data = []
    label = []
    for i in name:
        dataFile = open(path + '25-users/' + i + '.txt')
        data.append(np.array([eval(i) for i in dataFile.read().split()]).reshape(512, 14).T)
        labelFile = open(path + 'labels/' + i + '.lab')
        if labelFile.read() == 'Like':
            label.append(1)
        else:
            label.append(0)
    data = np.array(data)
    labels = np.array(label).T
    return data, labels

def load_SEED(folder_path, feature_name, frequency_band):
    '''
    :param folder_path: directory of ExtractedFeatures
    :param feature_name: feature name, for example 'de_LDS', 'asm_LDS' etc. Take de_LDS1 as an example: the demension is (62, 235, 5), 62 for 62 channels, 235 for 235 seconds and 5 for 5 different frequency bands.
    :param frequency_band: the input band name: 'delta', 'theta', 'alpha', 'beta', 'gamma'
    :return numpy array of data and labels
    '''
    frequency_idx = get_frequency_band_idx(frequency_band)
    labels = get_labels(os.path.join(folder_path, 'label.mat'))
    feature_vector_dict = {}
    label_dict = {}
    try:
        all_mat_file = os.walk(folder_path)
        skip_set = {'label.mat', 'readme.txt'}
        file_cnt = 0
        for path, dir_list, file_list in all_mat_file:
            for file_name in file_list:
                file_cnt += 1
                print('Currently process: {}, total progress: {}/{}'.format(file_name, file_cnt, len(file_list)))
                if file_name not in skip_set:
                    all_features_dict = loadmat(os.path.join(folder_path, file_name),
                                                     verify_compressed_data_integrity=False)
                    subject_name = file_name.split('.')[0]
                    feature_vector_trial_dict = {}
                    label_trial_dict = {}
                    for trials in range(1, 16):
                        feature_vector_list = []
                        label_list = []
                        cur_feature = all_features_dict[feature_name + str(trials)]
                        cur_feature = np.asarray(cur_feature[:, :, frequency_idx]).T  # dimensions: N * 62, N is the length of video
                        feature_vector_list.extend(_ for _ in cur_feature)
                        for _ in range(len(cur_feature)):
                            label_list.append(labels[trials - 1])
                        feature_vector_trial_dict[str(trials)] = feature_vector_list
                        label_trial_dict[str(trials)] = label_list
                    feature_vector_dict[subject_name] = feature_vector_trial_dict
                    label_dict[subject_name] = label_trial_dict
                else:
                    continue
    except FileNotFoundError as e:
        print('Loading Error: {}'.format(e))

    data = []
    labels = []
    for experiment in feature_vector_dict.keys():
        for trial in feature_vector_dict[experiment].keys():
            data.extend(feature_vector_dict[experiment][trial])
            labels.extend(label_dict[experiment][trial])
    data = np.array(data)
    labels = np.array(labels).T
    return data, labels



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="deeg",
    version="0.1.9",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "numpy>=1.18.0",
        "scipy>=1.5.4",
        "re>=2.2.1",
        "mne>=0.22.0",
        "pathlib>=1.0.1",
        "pytorch>=1.7.0"

    ],
    author="Jiaqi Li, Zhihao Zhao, Yiming Li",
    author_email="lijiaqi199609@sina.com, sethzhao506@berkeley.edu",
    description="package for EEG data analysis and deep learning",
    license="MIT",
    url="https://github.com/pypa/sampleproject",
    long_description=long_description,
    long_description_content_type="text/markdown"
)



import scipy
import numpy as np


def power_spectral_density(array, signal_frequency, window_size):
    """
    Calculate the power spectral density of given signal.
    We support both 1d array input and DEEG standard input (p,m,e). For the later format, we calculate PSD on dim "e" and return a (p,m) numpy array.

    input:
        array: numpy array. The temporal signal for feature extraction. 1d or with multiple dimensions
        signal_frequency: int. Original frequency of the signal
        window_size: int. Size of the window
    return
        PSD: tuple or numpy array, depends on the input shape
    """
    def cal_psd(s, sf, ws):
        freqs, psd= scipy.signal.welch(s, fs=sf, window='hanning', nperseg=ws, noverlap=None, nfft=None, detrend='constant', return_onesided=True, scaling='density', axis=-1)
        return psd
    if len(array.shape) == 1:
        return cal_psd(array, signal_frequency, window_size)
    else:
        return np.apply_along_axis(cal_psd, axis = -1, arr = array, sf = signal_frequency, ws = window_size)


def mean_absolute_deviation(array):
    """
    Calculate the mean_absolute_deviation on EEG data

    input:
        array: numpy array. The temporal signal for feature extraction. 1d or with multiple dimensions
    return
        de: int or numpy array, depends on the input shape
    """
    def cal_mad(s):
        return np.mean(np.abs(s - np.mean(s)))
    if len(array.shape) < 2:
        return cal_mad(array)
    else:
        return np.apply_along_axis(cal_mad, axis=-1, arr=array)


def skewness(array):
    """
    Calculate the feature "skewness" of the EEG signal
    We support both 1d array input and DEEG standard input (p,m,e). For the later format, we calculate SKEW on dim "e" and return a (p,m) numpy array.

    input:
        array: numpy array. The temporal signal for feature extraction. 1d or with multiple dimensions
    return
        SKEW: int or numpy array, depends on the input shape
    """
    def cal_skewness(s):
        n = len(s)
        ave1 = 0.0
        ave2 = 0.0
        ave3 = 0.0
        for x in s:
            ave1 += x
            ave2 += x ** 2
            ave3 += x ** 3
        ave1 /= n
        ave2 /= n
        ave3 /= n
        sigma = np.sqrt(ave2 - ave1 ** 2)
        return (ave3 - 3 * ave1 * sigma ** 2 - ave1 ** 3) / (sigma ** 3)
    if len(array.shape) == 1:
        return cal_skewness(array)
    else:
        return np.apply_along_axis(cal_skewness, axis=-1, arr=array)


def kurtosis(array):
    """
    Calculate the feature "kurtosis" of the EEG signal
    We support both 1d array input and DEEG standard input (p,m,e). For the later format, we calculate KURT on dim "e" and return a (p,m) numpy array.

    input:
        array: numpy array. The temporal signal for feature extraction. 1d or with multiple dimensions
    return
        KURT: int or numpy array, depends on the input shape
    """
    def cal_kurtosis(s):
        n = len(s)
        ave1 = 0.0
        ave2 = 0.0
        ave4 = 0.0
        for x in s:
            ave1 += x
            ave2 += x ** 2
            ave4 += x ** 4
        ave1 /= n
        ave2 /= n
        ave4 /= n
        sigma = np.sqrt(ave2 - ave1 ** 2)
        return ave4 / (sigma ** 4)
    if len(array.shape) == 1:
        return cal_kurtosis(array)
    else:
        return np.apply_along_axis(cal_kurtosis, axis=-1, arr=array)


def differential_entropy(array):
    """
    Calculate the feature "Differential Entropy" following the paper: "Differential Entropy Feature for EEG-based Vigilance Estimation", with hypothesis that processed signal is normal-distributed
    We support both 1d array input and DEEG standard input (p,m,e). For the later format, we calculate DE on dim "e" and return a (p,m) numpy array.
    input:
        array: numpy array. The temporal signal for feature extraction. 1d or with multiple dimensions
    return:
        de: int or numpy array, depends on the input shape
    """
    def cal_de(s):
        return (1/2)*np.log(2*np.pi*np.e*np.std(s)**2)
    if len(array.shape)<2:
        return cal_de(array)
    else:
        return np.apply_along_axis(cal_de, axis=-1, arr=array)


def rational_asymmetry(array, channel_name):
    """
    Calculate the rational asymmetry, of EEG singnal
    The differential entropy of the right brain channel is divided by that of the corresponding left brain channel

    input:
        array: numpy array. The temporal signal (>=2d) for feature extraction.
        channel_name: list or 1d array. The names of all the channels of the EEG signal
    return
        rasm: numpy array.
    """
    if len(channel_name) != array.shape[-2]:
        raise ValueError("The length of channel_name must be the same as the number of channels of the signal")
    DE = differential_entropy(array)
    def cal_rasm(de, cn):
        signal_DE = dict(zip(cn, de))
        y = []
        for key in signal_DE:
            if key == 'Fp1':
                y.append(signal_DE['Fp1'] / signal_DE['Fp2'])
            if key == 'AF7':
                y.append(signal_DE['AF7'] / signal_DE['AF8'])
            if key == 'AF3':
                y.append(signal_DE['AF3'] / signal_DE['AF4'])
            if key == 'F9':
                y.append(signal_DE['F9'] / signal_DE['F10'])
            if key == 'F7':
                y.append(signal_DE['F7'] / signal_DE['F8'])
            if key == 'F5':
                y.append(signal_DE['F5'] / signal_DE['F6'])
            if key == 'F3':
                y.append(signal_DE['F3'] / signal_DE['F4'])
            if key == 'F1':
                y.append(signal_DE['F1'] / signal_DE['F2'])
            if key == 'FT9':
                y.append(signal_DE['FT9'] / signal_DE['FT10'])
            if key == 'FT7':
                y.append(signal_DE['FT7'] / signal_DE['FT8'])
            if key == 'FC5':
                y.append(signal_DE['FC5'] / signal_DE['FC6'])
            if key == 'FC3':
                y.append(signal_DE['FC3'] / signal_DE['FC4'])
            if key == 'FC1':
                y.append(signal_DE['FC1'] / signal_DE['FC2'])
            if key == 'LPA':
                y.append(signal_DE['LPA'] / signal_DE['RPA'])
            if key == 'T7':
                y.append(signal_DE['T7'] / signal_DE['T8'])
            if key == 'C5':
                y.append(signal_DE['C5'] / signal_DE['C6'])
            if key == 'C3':
                y.append(signal_DE['C3'] / signal_DE['C4'])
            if key == 'C1':
                y.append(signal_DE['C1'] / signal_DE['C2'])
            if key == 'TP9':
                y.append(signal_DE['TP9'] / signal_DE['TP10'])
            if key == 'TP7':
                y.append(signal_DE['TP7'] / signal_DE['TP8'])
            if key == 'CP5':
                y.append(signal_DE['CP5'] / signal_DE['CP6'])
            if key == 'CP3':
                y.append(signal_DE['CP3'] / signal_DE['CP4'])
            if key == 'CP1':
                y.append(signal_DE['CP1'] / signal_DE['CP2'])
            if key == 'P9':
                y.append(signal_DE['P9'] / signal_DE['P10'])
            if key == 'P7':
                y.append(signal_DE['P7'] / signal_DE['P8'])
            if key == 'P5':
                y.append(signal_DE['P5'] / signal_DE['P6'])
            if key == 'P3':
                y.append(signal_DE['P3'] / signal_DE['P4'])
            if key == 'P1':
                y.append(signal_DE['P1'] / signal_DE['P2'])
            if key == 'PO7':
                y.append(signal_DE['PO7'] / signal_DE['PO8'])
            if key == 'PO3':
                y.append(signal_DE['PO3'] / signal_DE['PO4'])
            if key == 'O1':
                y.append(signal_DE['O1'] / signal_DE['O2'])
        return np.array(y)
    if len(DE.shape) == 1:
        return cal_rasm(DE, channel_name)
    else:
        return np.apply_along_axis(cal_rasm, axis=-1, arr=DE, cn=channel_name)


def differential_asymmetry(array, channel_name):
    """
    Calculate the differential asymmetry of EEG singnal
    The differential entropy of the right brain channel was subtracts from that of the corresponding left brain channel

    input:
        array: numpy array. The temporal signal (>=2d) for feature extraction
        channel_name: list or 1d array. The names of all the channels of the EEG signal
    return
        dasm: numpy array.
    """
    if len(channel_name) != array.shape[-2]:
        raise ValueError("The length of channel_name must be the same as the number of channels of the signal")
    DE = differential_entropy(array)
    def cal_dasm(de, cn):
        signal_DE = dict(zip(cn, de))
        y = []
        for key in signal_DE:
            if key == 'Fp1':
                y.append(signal_DE['Fp1'] - signal_DE['Fp2'])
            if key == 'AF7':
                y.append(signal_DE['AF7'] - signal_DE['AF8'])
            if key == 'AF3':
                y.append(signal_DE['AF3'] - signal_DE['AF4'])
            if key == 'F9':
                y.append(signal_DE['F9'] - signal_DE['F10'])
            if key == 'F7':
                y.append(signal_DE['F7'] - signal_DE['F8'])
            if key == 'F5':
                y.append(signal_DE['F5'] - signal_DE['F6'])
            if key == 'F3':
                y.append(signal_DE['F3'] - signal_DE['F4'])
            if key == 'F1':
                y.append(signal_DE['F1'] - signal_DE['F2'])
            if key == 'FT9':
                y.append(signal_DE['FT9'] - signal_DE['FT10'])
            if key == 'FT7':
                y.append(signal_DE['FT7'] - signal_DE['FT8'])
            if key == 'FC5':
                y.append(signal_DE['FC5'] - signal_DE['FC6'])
            if key == 'FC3':
                y.append(signal_DE['FC3'] - signal_DE['FC4'])
            if key == 'FC1':
                y.append(signal_DE['FC1'] - signal_DE['FC2'])
            if key == 'LPA':
                y.append(signal_DE['LPA'] - signal_DE['RPA'])
            if key == 'T7':
                y.append(signal_DE['T7'] - signal_DE['T8'])
            if key == 'C5':
                y.append(signal_DE['C5'] - signal_DE['C6'])
            if key == 'C3':
                y.append(signal_DE['C3'] - signal_DE['C4'])
            if key == 'C1':
                y.append(signal_DE['C1'] - signal_DE['C2'])
            if key == 'TP9':
                y.append(signal_DE['TP9'] - signal_DE['TP10'])
            if key == 'TP7':
                y.append(signal_DE['TP7'] - signal_DE['TP8'])
            if key == 'CP5':
                y.append(signal_DE['CP5'] - signal_DE['CP6'])
            if key == 'CP3':
                y.append(signal_DE['CP3'] - signal_DE['CP4'])
            if key == 'CP1':
                y.append(signal_DE['CP1'] - signal_DE['CP2'])
            if key == 'P9':
                y.append(signal_DE['P9'] - signal_DE['P10'])
            if key == 'P7':
                y.append(signal_DE['P7'] - signal_DE['P8'])
            if key == 'P5':
                y.append(signal_DE['P5'] - signal_DE['P6'])
            if key == 'P3':
                y.append(signal_DE['P3'] - signal_DE['P4'])
            if key == 'P1':
                y.append(signal_DE['P1'] - signal_DE['P2'])
            if key == 'PO7':
                y.append(signal_DE['PO7'] - signal_DE['PO8'])
            if key == 'PO3':
                y.append(signal_DE['PO3'] - signal_DE['PO4'])
            if key == 'O1':
                y.append(signal_DE['O1'] - signal_DE['O2'])
        return np.array(y)
    if len(DE.shape) == 1:
        return cal_dasm(DE, channel_name)
    else:
        return np.apply_along_axis(cal_dasm, axis=-1, arr=DE, cn=channel_name)


def differential_caudality(array, channel_name):
    """
    Calculate the differential caudality of EEG singnal
    The differential entropy of the posterior brain channel was subtracts from that of the corresponding frontal brain channel

    input:
        array: numpy array. The temporal signal (>=2d) for feature extraction.
        channel_name: list or 1d array. The names of all the channels of the EEG signal
    return
        dcau: numpy array.
    """
    if len(channel_name) != array.shape[-2]:
        raise ValueError("The length of channel_name must be the same as the number of channels of the signal")
    DE = differential_entropy(array)

    def cal_dcau(de, cn):
        signal_DE = dict(zip(cn, de))
        y = []
        for key in signal_DE:
            if key == 'Nz':
                y.append(signal_DE['Nz'] - signal_DE['Iz'])
            if key == 'Fp1':
                y.append(signal_DE['Fp1'] - signal_DE['O1'])
            if key == 'Fpz':
                y.append(signal_DE['Fpz'] - signal_DE['Oz'])
            if key == 'Fp2':
                y.append(signal_DE['Fp2'] - signal_DE['O2'])
            if key == 'AF7':
                y.append(signal_DE['AF7'] - signal_DE['PO7'])
            if key == 'AF3':
                y.append(signal_DE['AF3'] - signal_DE['PO3'])
            if key == 'AFz':
                y.append(signal_DE['AFz'] - signal_DE['POz'])
            if key == 'AF4':
                y.append(signal_DE['AF4'] - signal_DE['PO4'])
            if key == 'AF8':
                y.append(signal_DE['AF8'] - signal_DE['PO8'])
            if key == 'F9':
                y.append(signal_DE['F9'] - signal_DE['P9'])
            if key == 'F7':
                y.append(signal_DE['F7'] - signal_DE['P7'])
            if key == 'F5':
                y.append(signal_DE['F5'] - signal_DE['P5'])
            if key == 'F3':
                y.append(signal_DE['F3'] - signal_DE['P3'])
            if key == 'F1':
                y.append(signal_DE['F1'] - signal_DE['P1'])
            if key == 'Fz':
                y.append(signal_DE['Fz'] - signal_DE['Pz'])
            if key == 'F2':
                y.append(signal_DE['F2'] - signal_DE['P2'])
            if key == 'F4':
                y.append(signal_DE['F4'] - signal_DE['P4'])
            if key == 'F6':
                y.append(signal_DE['F6'] - signal_DE['P6'])
            if key == 'F8':
                y.append(signal_DE['F8'] - signal_DE['P8'])
            if key == 'F10':
                y.append(signal_DE['F10'] - signal_DE['P10'])
            if key == 'FT9':
                y.append(signal_DE['FT9'] - signal_DE['TP9'])
            if key == 'FT7':
                y.append(signal_DE['FT7'] - signal_DE['TP7'])
            if key == 'FC5':
                y.append(signal_DE['FC5'] - signal_DE['CP5'])
            if key == 'FC3':
                y.append(signal_DE['FC3'] - signal_DE['CP3'])
            if key == 'FC1':
                y.append(signal_DE['FC1'] - signal_DE['CP1'])
            if key == 'FCz':
                y.append(signal_DE['FCz'] - signal_DE['CPz'])
            if key == 'FC2':
                y.append(signal_DE['FC2'] - signal_DE['CP2'])
            if key == 'FC4':
                y.append(signal_DE['FC4'] - signal_DE['CP4'])
            if key == 'FC6':
                y.append(signal_DE['FC6'] - signal_DE['CP6'])
            if key == 'FT8':
                y.append(signal_DE['FT8'] - signal_DE['TP8'])
            if key == 'FT10':
                y.append(signal_DE['FT10'] - signal_DE['TP10'])
        return np.array(y)

    if len(DE.shape) == 1:
        return cal_dcau(DE, channel_name)
    else:
        return np.apply_along_axis(cal_dcau, axis=-1, arr=DE, cn=channel_name)


def cal_eeg_features(array, sf, ws, channel_name=None):
    """
    Calculate all EEG features. If "channel_name" is not None, also include multi-channel features
    input:
        array: numpy array. The temporal signal for feature extraction. 1d or with multiple dimensions
        sf: int or float. Signal frequency for PSD calculation.
        wz: int or float. Window size for PSD calculation.
        channel_name: list or 1d array. The names of all the channels of the EEG signal.
    return
        feature_array: numpy array. Calculated features with the same shape as input array.
        feature_list: list of str. The order of EEG features, corresponding to "feature array".
    """
    feature_list = ["PSD", "mean_absolute_difference", "skewness", "kurtosis", "DE"]
    if len(array.shape)<2:
        # If input is a single signal, multi-channel features are excluded
        psd = power_spectral_density(array, sf, ws)
        mad = mean_absolute_deviation(array)
        skw = skewness(array)
        krt = kurtosis(array)
        de = differential_entropy(array)
        return np.append(psd, [mad,skw,krt,de]), feature_list
    else:
        # add new axis for numeric features to concatenate with vectorized feature
        psd = power_spectral_density(array, sf, ws)
        mad = mean_absolute_deviation(array)[..., np.newaxis]
        skw = skewness(array)[..., np.newaxis]
        krt = kurtosis(array)[..., np.newaxis]
        de = differential_entropy(array)[..., np.newaxis]
        if channel_name is not None:
            feature_list.append(["RASM", "DASM", "DCAU"])
            rasm = rational_asymmetry(array, channel_name)
            dasm = differential_asymmetry(array, channel_name)
            dcau = differential_caudality(array, channel_name)
            return np.concatenate([psd,mad,skw,krt,de,rasm,dasm,dcau],axis=-1), feature_list
        else:
            return np.concatenate([psd,mad,skw,krt,de],axis=-1), feature_list


